/*
 * if.c:
 * SLIP interface side of vpn-udp.
 *
 * Copyright (c) 2001 Chris Lightfoot. All rights reserved.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 */

static const char rcsid[] = "$Id: if.c,v 1.1.1.1 2001/08/30 18:24:39 chris Exp $";

#include <sys/types.h>

#include <errno.h>
#include <fcntl.h>
#include <pty.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

#include "vpn-udp.h"

extern int verbose; /* in main.c */

/* File descriptor on interface (master side of the PTY). */
int fd_if;

/* Name of terminal. */
char *tty_name;

/* Name of network interface. */
char *if_name;

/* create_interface:
 * Open a PTY and make it into a SLIP interface. Returns 0 on success or
 * -1 on failure. */
int create_interface(void) {
    /* Obtain pty. */
    int fd_pty2;
    int res;
    char *cmdline;

    if (openpty(&fd_if, &fd_pty2, NULL, NULL, NULL) == -1) {
        perror(PROGNAME ": openpty");
        return 0;
    }

    tty_name = strdup(ttyname(fd_pty2));

    if (verbose)
        print_log(LOG_ERR, "using TTY device `%s'", tty_name);

    /* Now execute slattach to make this into a SLIP interface. */
    cmdline = (char*)malloc(24 + strlen(tty_name));
    if (verbose)
        sprintf(cmdline, "slattach -e -p slip -d %s", tty_name);
    else
        sprintf(cmdline, "slattach -e -p slip %s", tty_name);

    if (verbose)
        print_log(LOG_ERR, "executing %s", cmdline);
    
    errno = 0;
    res = system(cmdline);
    free(cmdline);
    if ((res == 127 && errno != 0) || res == -1) {
        print_log(LOG_ERR, "system(`%s'): %s", cmdline, strerror(errno));
        close(fd_if);
        close(fd_pty2);
        return 0;
    }
    
    /* Obtain the network interface name. */
    if_name = (char*)malloc(256);
    if (ioctl(fd_pty2, SIOCGIFNAME, if_name) == -1) {
        perror(PROGNAME ": ioctl(SIOCGIFNAME)");
        close(fd_if);
        close(fd_pty2);
        return 0;
    }

    print_log(LOG_ERR, "network interface is `%s'", if_name);

    /* At this stage we should run some program to configure the interface. */

    return 1;
}

unsigned char *slip_pkt;

#define CH_ESC      ((unsigned char)0333)
#define CH_END      ((unsigned char)0300)

/* write_to_if:
 * Write a packet to the local interface. */
int write_to_if(unsigned char *pkt, int length) {
    unsigned char *p, *q;
    if (!slip_pkt) slip_pkt = (unsigned char*)malloc(BUFLEN);
    *slip_pkt = CH_END;
    for (p = pkt, q = slip_pkt + 1; p < pkt + length; ++p, ++q) {
        switch (*p) {
            case CH_END:
                *q++ = CH_ESC;
                *q   = 0334;
                break;
                
            case CH_ESC:
                *q++ = CH_ESC;
                *q   = 0335;
                break;

            default:
                *q = *p;
                break;
        }
    }
    *q++ = CH_END;

    if (write(fd_if, slip_pkt, q - slip_pkt) == -1) {
        print_log(LOG_ERR, "write(local interface): %m");
        return 0;
    }

    return 1;
}

/* read_from_if:
 * Read a packet, or part of one, from the local interface. */
unsigned char *if_buff;
unsigned char *if_ptr;
int read_from_if(void) {
    ssize_t n;
    unsigned char *p;
        
    if (!if_buff) if_buff = if_ptr = (unsigned char*)malloc(BUFLEN);
 
    while ((n = read(fd_if, if_ptr, if_buff + BUFLEN - if_ptr)) == -1 && errno == EINTR);

    if (n == -1) {
        if (errno == EIO) {
            /* We get EIO when the interface is down; ignore it. */
            fprintf(stderr, "#");
            return 1;
        } else {
            print_log(LOG_ERR, "read(local interface): %m");
            return 0;
        }
    } else if (n == 0) {
        if (verbose)
            print_log(LOG_ERR, "local interface closed");
        return 0;
    }

    if_ptr += n;

    /* Now look for packets at the front of the buffer. SLIP packets are ended
     * by an END character, and often begin that way as well. END characters
     * within a packet are escaped using ESC 0334 and escape characters with
     * ESC 335. */
    p = memchr(if_buff, CH_END, if_ptr - if_buff);
    do {
        /* Skip leading END characters. */
        for (; p < if_ptr && *p == CH_END; ++p);
        if (p < if_ptr) memmove(if_buff, p, if_buff + BUFLEN - p);
        if_ptr -= p - if_buff;

        if ((p = memchr(if_buff, CH_END, if_ptr - if_buff))) {
            unsigned char *q, *r;
            int bogon = 0;

            if (!slip_pkt) slip_pkt = (unsigned char*)malloc(BUFLEN);
            for (q = if_buff, r = slip_pkt; q < p && !bogon; ++q, ++r) {
                if (*q == CH_ESC) {
                    if (q == p - 1) {
                        print_log(LOG_ERR, "malformed trailing escape in local packet length %d", p - if_buff);
                        bogon = 1;
                        break;
                    }
                    switch (*(q + 1)) {
                        case 0334:
                            *r = CH_END;
                            break;

                        case 0335:
                            *r = CH_ESC;
                            break;

                        default:
                            print_log(LOG_ERR, "malformed escape in local packet length %d", p - if_buff);
                            bogon = 1;
                            break;
                    }
                    ++q;
                } else *r = *q;
            }

            if (!bogon) write_to_tunnel(slip_pkt, r - slip_pkt);

            /* Now consume the first p - if_buff bytes. */
            memmove(if_buff, p, if_buff + BUFLEN - p);
            if_ptr -= p - if_buff;
        }
    } while ((p = memchr(if_buff, CH_END, if_ptr - if_buff)));

    return 1;
}

