/*
 * logprint.c:
 * Functions for logging messages to the syslog, and optionally standard error
 * as well.
 *
 * Copyright (c) 2001 Chris Lightfoot. All rights reserved.
 *
 */

static const char rcsid[] = "$Id: logprint.c,v 1.1.1.1 2001/08/30 18:24:39 chris Exp $";

#include <sys/types.h>

#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>

#include <sys/utsname.h>

#include "vpn-udp.h"

/* verrprintf:
 * Returns a malloc'd string with the appropriate arguments printed into it. */
#define GUESS 32
static char *verrprintf(const char *fmt, va_list ap) {
    size_t l = GUESS;
    int n;
    char *s, *e = strerror(errno), *t;
    size_t le = strlen(e);
    const char *p, *q;
    
    if (!(s = malloc(l))) return NULL;
    memset(s, 0, l);

    /* First, we need to substitute errors into the string. */
    for (p = fmt, q = strstr(p, "%m"); q; p = q, q = strstr(p, "%m")) {
        if ((q - p + le) > l - strlen(s)) {
            l = (q - p + le) - strlen(s) + 1;
            s = realloc(s, l);
        }
        strncat(s, p, q - p);
        strcat(s, e);
        q += 2;
    }
    if (strlen(p) > (l - strlen(s))) {
        l = strlen(s) + strlen(p) + 1;
        s = realloc(s, l);
    }

    strcat(s, p);

    /* Now, need to use the generated string as a format string to pass to vsnprintf. */
    while (1) {
        t = (char*)malloc(n = l);
        n = vsnprintf(t, n, s, ap);
        if (n > -1 && n < l) {
            free(s);
            return t;
        }
        if (n > -1) l = n + 1;  /* Cope with vsnprintf returning either the required amount of space, or -1 to indicate "need more". */
        else l *= 2;
        if (!(t = realloc(t, l))) return NULL;
    }
}

/* print_log:
 * Because many systems do not have LOG_PERROR, we use a custom function to
 * write an error to the system log, and optionally to standard error as
 * well. */
void print_log(int priority, const char *fmt, ...) {
    char *s;
    va_list ap;
    extern int log_stderr;      /* in main.c */
    static char *name;

    va_start(ap, fmt);
    s = verrprintf(fmt, ap);
    va_end(ap);
    syslog(priority, "%s", s);
    if (log_stderr) {
        /* It's useful to print the identity of the machine that's printing a
         * diagnostic. */
        if (!name) {
            struct utsname uts;
            uname(&uts);
            name = strdup(uts.nodename);
        }
        fprintf(stderr, PROGNAME"(%s): %s\n", name, s);
    }
    free(s);
}
