/*
 * tunnel.c:
 * Tunnel side of vpn-udp.
 *
 * Copyright (c) 2001 Chris Lightfoot. All rights reserved.
 * Email: chris@ex-parrot.com; WWW: http://www.ex-parrot.com/~chris/
 *
 */

static const char rcsid[] = "$Id: tunnel.c,v 1.4 2001/08/31 20:41:11 chris Exp $";

#include <sys/types.h>

#include <errno.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>

#include <arpa/inet.h>
#include <netinet/in.h>

#include <sys/ioctl.h>
#include <sys/socket.h>

#include "vpn-udp.h"

extern int verbose; /* in main.c */

/* File descriptor on network. */
int fd_tunnel;

/* Addresses of remote and local ends of link. The program can operate in a
 * mode where it guesses the remote port once a packet has been received; in
 * this mode, remote_port is left equal to 0 until a packet is received. */
struct in_addr remote_addr;
short local_port, remote_port;

/* Timestamps of last packets sent/received, used to figure out whether the
 * link is still available. */
time_t last_sent, last_received;

/* connect_to_remote:
 * Connect the tunnel to the remote host. */
int connect_to_remote(void) {
    struct sockaddr_in sin = {0};
    sin.sin_family = AF_INET;
    memcpy(&sin.sin_addr, &remote_addr, sizeof(remote_addr));
    sin.sin_port = htons((short)remote_port);
    if (connect(fd_tunnel, &sin, sizeof(sin)) == -1) {
        print_log(LOG_ERR, "connect: %m");
        return 0;
    }
    if (verbose)
        print_log(LOG_ERR, "connected to %s:%d", inet_ntoa(remote_addr), remote_port);
    return 1;
}

/* write_to_tunnel:
 * Send a packet to the remote host. */
unsigned char *net_pkt;

int write_to_tunnel(unsigned char *packet, int length) {
    char rbuf[17];
    int padlen, padlen2, netlen = 0;
    short s;

    /* Packets are sent in the following format:
     *
     * : 1 byte:             :2 bytes:
     * +-------+---- - - ----+-------+---- - - ----+---- - - ----+
     * |padding| / padding / | packet| /  packet / | / padding / |
     * | length| / / / / / / | length| / / data  / | / / / / / / |
     * +-------+---- - - ----+-------+---- - - ----+---- - - ----+
     *
     * where the padding is randomly-generated, and the lengths are sent as
     * shorts in network order. We use up to 16 bytes of padding ahead and
     * behind the packet itself. The whole packet must be a multiple of 8
     * bytes in length.
     *
     * Packets on the wire are encrypted using Blowfish. */
    if (!net_pkt) net_pkt = (unsigned char*)malloc(BUFLEN);
    if (!get_random_data(rbuf, 17)) return 0;
    padlen = (rbuf[0] & 0xf) + 1;
    net_pkt[0] = (unsigned char)padlen;             netlen += 1;
    memcpy(net_pkt + netlen, rbuf + 1, padlen);     netlen += padlen;

    s = htons((short)length);
    memcpy(net_pkt + netlen, &s, 2);                netlen += 2;
    memcpy(net_pkt + netlen, packet, length);       netlen += length;

    /* Ensure that trailing padding makes packet up to a multiple of 8 bytes. */
    if (!get_random_data(rbuf, 17)) return 0;
    padlen2 = (rbuf[0] & 7) + 1;
    padlen2 += 8 - ((netlen + padlen2) & 7);
    memcpy(net_pkt + netlen, rbuf + 1, padlen2);    netlen += padlen2;

    encrypt_packet(net_pkt, netlen);

    if (write(fd_tunnel, net_pkt, netlen) == -1) {
        print_log(LOG_ERR, "write(tunnel): %m");
        return 0;
    }

    last_sent = time(NULL);
    
    return 1;
}

/* read_from_tunnel:
 * Read a packet from the remote host. */
int read_from_tunnel(void) {
    ssize_t n;
    short s;
    int padlen, pktlen;
    struct sockaddr_in sin;
    int sinlen = sizeof(sin);
 
    if (!net_pkt) net_pkt = (unsigned char*)malloc(BUFLEN);
    
    while ((n = recvfrom(fd_tunnel, net_pkt, BUFLEN, 0, &sin, &sinlen)) == -1 && errno == EINTR);

    if (n == -1) {
        print_log(LOG_ERR, "read(tunnel): %m");
        if (errno != ECONNREFUSED) return 0;    /* may get this waiting for remote end to come up */
        else return 1;
    }

    /* Is this from the right place? */
    if (memcmp(&sin.sin_addr, &remote_addr, sizeof(remote_addr))
        || (remote_port && htons((short)remote_port) != sin.sin_port)) {
        print_log(LOG_ERR, "received packet from %s:%d, which is not our peer",
                inet_ntoa(sin.sin_addr), ntohs(sin.sin_port));
        return 1;
    }

    if ((n & 7) != 0) {
        print_log(LOG_ERR, "packet length not a multiple of 8 bytes in tunnel packet");
        return 1;
    }
    
    decrypt_packet(net_pkt, n);

    /* Get padding length. */
    padlen = (int)net_pkt[0];
    if (padlen > n - 1) {
        print_log(LOG_ERR, "bogus padding length in tunnel packet length %d", n);
        return 1;
    }

    /* Get packet length. */
    memcpy(&s, net_pkt + 1 + padlen, 2);
    pktlen = ntohs(s);
    if (1 + padlen + 2 + pktlen > n) {
        print_log(LOG_ERR, "bogus packet length in tunnel packet length %d", n);
        return 1;
    }

    /* If we're supposed to figure out the remote port, we know it now;
     * connect. */
    if (!remote_port) {
        remote_port = ntohs(sin.sin_port);
        if (!connect_to_remote()) return 0;
    }
    
    last_received = time(NULL);

    /* Send the packet on into the SLIP interface. Packets of length 0 are
     * internal keepalive packets, not real IP packets. */
    if (pktlen > 0) return write_to_if(net_pkt + 1 + padlen + 2, pktlen);
    else return 1;
}

